/*
    Fast matrix multiply contest entry
    by Edward Keyes, April 10, 1997
    for Edelman's 18.337 Homework 5

    This matrix multiply routine is optimized for the Sun UltraSPARC processor
    architecture used in the Xolas machine.  This processor contains 32 floating
    point registers and two floating point execution units.  The level-1 cache
    works in 16-byte blocks, so it isn't of much help here.

    The philosophy behind the program is to minimize memory access by using
    the register variables as much as possible.  The routine calculates a 4x4
    block of answers in one pass, performing 16 multiplies and adds per 8 memory
    accesses, a ratio 4 times better than the naive routine.  An extra 8 registers
    are used as temporary storage for pipelining the calculations, and for
    interleaving the memory loads between the calculations.

    Although it cuts into performance a little bit (for a 61x61 matrix multiply,
    about 9% extra work is done), the boundary cases are not treated differently.
    Answers are calculated with dummy values, but just never stored.  This is
    something that can easily be added, though, to gain those few percent.
*/

#include <stdio.h>

void mul_mfmf_mf(int size1, int size2, int size3,
      double *A, double *B, double *C)
/*
   Our matrix multiply routine.  No idea why the C array was declared constant.
   We have used the convention that C=A*B, with A having size1 rows, size2
   columns, and B having size2 rows and size3 columns.  This convention was
   inferred from the given naive routine.

   Everything is declared register, just to wake the compiler up and encourage
   it not to preserve registers.
*/
{
   register double c00, c01, c02, c03, c10, c11, c12, c13;
   register double c20, c21, c22, c23, c30, c31, c32, c33;
   register double ax0, ax1, ax2, ax3;
   register double bx0, bx1, bx2, bx3;

   register double *baseC, *baseA, *baseB;
   register int numr, numc, numi;
   register int i,j,k;
   register int border1,border2;
   register int s1,s2,s3;
   register int mod1,mod2,mod3;

   numr=(size1+3)/4;   /* round to next highest multiple of 4 */
   numc=(size3+3)/4;
   mod1=size1 % 4;
   mod2=size2 % 2;
   mod3=size3 % 4;
   if (mod1==0) mod1=4;
   if (mod3==0) mod3=4;
   numi=size2/2;
   s1=size1; s2=size2; s3=size3; /* just copy to registers */
   i=0;
   while (i<numr) { /* loop over rows of C */
      if (i==numr-1)
         border1=1;
      else border1=0;
      border2=0;
      j=0;
      baseC=C+i*s3*4;
      while (j<numc) { /* loop over columns of C */
         if (j==numc-1)
            border2=1;
         k=0;
         baseA=A+i*4*s2;
         baseB=B+j*4;
         j++; /* increment early for the pipeline */
         c00=baseC[0];
         c01=baseC[1];
         c02=baseC[2];
         c03=baseC[3];
         baseC+=s3;
         c10=baseC[0];
         c11=baseC[1];
         c12=baseC[2];
         c13=baseC[3];
         baseC+=s3;
         c20=baseC[0];
         c21=baseC[1];
         c22=baseC[2];
         c23=baseC[3];
         baseC+=s3;
         c30=baseC[0];
         c31=baseC[1];
         c32=baseC[2];
         c33=baseC[3];
         while (k<numi) { /* do the calculations */
            k++;
            ax0=baseA[0];
            ax1=baseA[s2];
            ax2=baseA[s2+s2];
            ax3=baseA[s2+s2+s2];
            bx0=baseB[0];
            bx1=baseB[1];
            bx2=baseB[2];
            bx3=baseB[3];
            baseA+=1;
            baseB+=s3;
            c00+=ax0*bx0;
            c01+=ax0*bx1;
            c02+=ax0*bx2;
            c03+=ax0*bx3;
            c10+=ax1*bx0;
            c11+=ax1*bx1;
            c12+=ax1*bx2;
            c13+=ax1*bx3;
            c20+=ax2*bx0;
            c21+=ax2*bx1;
            c22+=ax2*bx2;
            c23+=ax2*bx3;
            c30+=ax3*bx0;
            c31+=ax3*bx1;
            c32+=ax3*bx2;
            c33+=ax3*bx3;
            ax0=baseA[0];
            ax1=baseA[s2];
            ax2=baseA[s2+s2];
            ax3=baseA[s2+s2+s2];
            bx0=baseB[0];
            bx1=baseB[1];
            bx2=baseB[2];
            bx3=baseB[3];
            baseA+=1;
            baseB+=s3;
            c00+=ax0*bx0;
            c01+=ax0*bx1;
            c02+=ax0*bx2;
            c03+=ax0*bx3;
            c10+=ax1*bx0;
            c11+=ax1*bx1;
            c12+=ax1*bx2;
            c13+=ax1*bx3;
            c20+=ax2*bx0;
            c21+=ax2*bx1;
            c22+=ax2*bx2;
            c23+=ax2*bx3;
            c30+=ax3*bx0;
            c31+=ax3*bx1;
            c32+=ax3*bx2;
            c33+=ax3*bx3;
         }
         if (mod2==1) {
            ax0=baseA[0];
            ax1=baseA[s2];
            ax2=baseA[s2+s2];
            ax3=baseA[s2+s2+s2];
            bx0=baseB[0];
            bx1=baseB[1];
            bx2=baseB[2];
            bx3=baseB[3];
            baseA+=1;
            baseB+=s3;
            c00+=ax0*bx0;
            c01+=ax0*bx1;
            c02+=ax0*bx2;
            c03+=ax0*bx3;
            c10+=ax1*bx0;
            c11+=ax1*bx1;
            c12+=ax1*bx2;
            c13+=ax1*bx3;
            c20+=ax2*bx0;
            c21+=ax2*bx1;
            c22+=ax2*bx2;
            c23+=ax2*bx3;
            c30+=ax3*bx0;
            c31+=ax3*bx1;
            c32+=ax3*bx2;
            c33+=ax3*bx3;
         }
         if ((!(border1+border2)) || (mod1+mod3==8)) {
            baseC[0]=c30;
            baseC[1]=c31;
            baseC[2]=c32;
            baseC[3]=c33;
            baseC-=s3;
            baseC[0]=c20;
            baseC[1]=c21;
            baseC[2]=c22;
            baseC[3]=c23;
            baseC-=s3;
            baseC[0]=c10;
            baseC[1]=c11;
            baseC[2]=c12;
            baseC[3]=c13;
            baseC-=s3;
            baseC[0]=c00;
            baseC[1]=c01;
            baseC[2]=c02;
            baseC[3]=c03;
         }
         else if (border1-border2==1) {
            if (mod1>3) {
               baseC[0]=c30;
               baseC[1]=c31;
               baseC[2]=c32;
               baseC[3]=c33;
            }
            baseC-=s3;
            if (mod1>2) {
               baseC[0]=c20;
               baseC[1]=c21;
               baseC[2]=c22;
               baseC[3]=c23;
            }
            baseC-=s3;
            if (mod1>1) {
               baseC[0]=c10;
               baseC[1]=c11;
               baseC[2]=c12;
               baseC[3]=c13;
            }
            baseC-=s3;
            baseC[0]=c00;
            baseC[1]=c01;
            baseC[2]=c02;
            baseC[3]=c03;
         }
         else if (border2-border1==1) {
            baseC[0]=c30;
            if (mod3>1) baseC[1]=c31;
            if (mod3>2) baseC[2]=c32;
            if (mod3>3) baseC[3]=c33;
            baseC-=s3;
            baseC[0]=c20;
            if (mod3>1) baseC[1]=c21;
            if (mod3>2) baseC[2]=c22;
            if (mod3>3) baseC[3]=c23;
            baseC-=s3;
            baseC[0]=c10;
            if (mod3>1) baseC[1]=c11;
            if (mod3>2) baseC[2]=c12;
            if (mod3>3) baseC[3]=c13;
            baseC-=s3;
            baseC[0]=c00;
            if (mod3>1) baseC[1]=c01;
            if (mod3>2) baseC[2]=c02;
            if (mod3>3) baseC[3]=c03;
         }
         else if (border1+border2==2) {
            if (mod1>3) {
               baseC[0]=c30;
               if (mod3>1) baseC[1]=c31;
               if (mod3>2) baseC[2]=c32;
               if (mod3>3) baseC[3]=c33;
            }
            baseC-=s3;
            if (mod1>2) {
               baseC[0]=c20;
               if (mod3>1) baseC[1]=c21;
               if (mod3>2) baseC[2]=c22;
               if (mod3>3) baseC[3]=c23;
            }
            baseC-=s3;
            if (mod1>1) {
               baseC[0]=c10;
               if (mod3>1) baseC[1]=c11;
               if (mod3>2) baseC[2]=c12;
               if (mod3>3) baseC[3]=c13;
            }
            baseC-=s3;
            baseC[0]=c00;
            if (mod3>1) baseC[1]=c01;
            if (mod3>2) baseC[2]=c02;
            if (mod3>3) baseC[3]=c03;
         }
         baseC+=4; /* next group of results */
      }
      i++;
   }
}
